package org.codehaus.groovy.grails.web.pages

import grails.util.GrailsUtil
import grails.util.GrailsWebUtil

import org.codehaus.groovy.grails.commons.GrailsApplication
import org.codehaus.groovy.grails.commons.GrailsClass
import org.codehaus.groovy.grails.plugins.MockGrailsPluginManager;
import org.codehaus.groovy.grails.support.MockStringResourceLoader
import org.springframework.core.io.ByteArrayResource
import org.springframework.core.io.UrlResource
import org.springframework.mock.web.MockServletContext
import org.springframework.web.context.request.RequestContextHolder

class GroovyPagesTemplateEngineTests extends GroovyTestCase {

    void testCommentAtEndOfTemplate() {
        def webRequest = GrailsWebUtil.bindMockWebRequest()
        def request = webRequest.request
        request.addParameter("showSource", "true")

        System.setProperty("grails.env", "development")
        assert GrailsUtil.isDevelopmentEnv()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        // It is important that the template ends with the comment. Whitespace or anything else after
        // the comment will not trigger the problem.  See GRAILS-1737
        def pageSource = "<html><body></body></html><%-- should not be in the output --%>"

        def t = gpte.createTemplate(pageSource, "comment_test")
        def w = t.make()

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertTrue(sw.toString().indexOf("should not be in the output") == -1)
    }

    void testShowSourceParameter() {
        try {
            def webRequest = GrailsWebUtil.bindMockWebRequest()
            def request = webRequest.request
            request.addParameter("showSource", "true")

            System.setProperty("grails.env", "development")
            assert GrailsUtil.isDevelopmentEnv()

            def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
            gpte.afterPropertiesSet()

            def t = gpte.createTemplate("<%='hello'%>", "hello_test")
            def w = t.make()

            def sw = new StringWriter()
            def pw = new PrintWriter(sw)

            w.writeTo(pw)

            assertTrue(sw.toString().indexOf(GroovyPage.OUT_STATEMENT + ".print('hello')") > -1)

        }
        finally {
            System.setProperty("grails.env", "")
        }
    }

    void testEstablishNameForResource() {
        def res = new UrlResource("http://grails.org/some.path/foo.gsp")

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        assertEquals "some_path_foo_gsp", gpte.establishPageName(res, null)
    }

    void testCreateTemplateFromCurrentRequest2() {
        def webRequest = GrailsWebUtil.bindMockWebRequest()


        def uri1 = "/another"
        assertNotNull(webRequest.request)
        webRequest.request.requestURI = "/another"
        webRequest.request.servletPath = "/another"

        def rl = new MockStringResourceLoader()
        rl.registerMockResource(uri1, "<%='success 2'%>")


        def gpte = new GroovyPagesTemplateEngine(new MockServletContext(rl))
        gpte.afterPropertiesSet()

        gpte.groovyPageLocator.addResourceLoader(rl)

        def t = gpte.createTemplate()
        def w = t.make()

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertEquals "success 2", sw.toString()
    }

    void testCreateTemplateFromCurrentRequest1() {
        def webRequest = GrailsWebUtil.bindMockWebRequest()

        def uri1 = "/somedir/myview"
        assertNotNull(webRequest.request)
        webRequest.request.requestURI = uri1
        webRequest.request.servletPath = uri1

        def uri2 = "/another"

        def rl = new MockStringResourceLoader()
        rl.registerMockResource(uri1, "<%='success 1'%>")
        rl.registerMockResource(uri2, "<%='success 2'%>")

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext(rl))
        gpte.afterPropertiesSet()

        gpte.groovyPageLocator.addResourceLoader(rl)

        def t = gpte.createTemplate()
        def w = t.make()

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertEquals "success 1", sw.toString()
    }

    void testCreateTemplateFromResource() {
        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        def t = gpte.createTemplate(new ByteArrayResource("<%='hello'%>".bytes))
        def w = t.make()

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertEquals "hello", sw.toString()
    }

    void testNestingGroovyExpressionInAttribute() {

        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        def src = '''<g:actionSubmit onclick="return confirm('${message}')"/>'''
        def t = gpte.createTemplate(src, "hello_test")

        def w = t.make(message: 'Are You Sure')

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertEquals '''<g:actionSubmit onclick="return confirm('Are You Sure')"></g:actionSubmit>''', sw.toString()
    }
    
    private GrailsApplication createMockGrailsApplication(ConfigObject config = null) {
        if(config==null) {
            config=new ConfigObject()
            config.put(GroovyPageParser.CONFIG_PROPERTY_GSP_KEEPGENERATED_DIR, System.getProperty("java.io.tmpdir"))
        }
        [getMainContext: { ->  null},  getConfig: { ->  config} , getFlatConfig: { -> config.flatten() } , getArtefacts: { String artefactType -> [] as GrailsClass[] }, getArtefactByLogicalPropertyName: { String type, String logicalName ->  null} ] as GrailsApplication
    }

    void testParsingNestedCurlyBraces() {
        // GRAILS-7915
        //if(notYetImplemented()) return

        def webRequest = GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.grailsApplication = createMockGrailsApplication()
        gpte.afterPropertiesSet()

        def src = '${people.collect {it.firstName}}'
        def t = gpte.createTemplate(src, "hello_test")

        def people = [[firstName: 'Peter', lastName: 'Gabriel'], [firstName: 'Phil', lastName: 'Collins']]
        def w = t.make(people: people)

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertEquals "[Peter, Phil]", sw.toString()

        src = '''
        <g:if env="production">
        </g:if>

        <script type="text/javascript">
        try {
        <g:if test="${title == 'Selling England By The Pound'}">
        ${bandName}
        </g:if>
        } catch( err ) {}
        </script>
'''

        gpte.createTemplate(src, "hello_test")
        t = gpte.createTemplate(src, "hello_test")
        w = t.make(bandName: 'Genesis', title: 'Selling England By The Pound')

        sw = new StringWriter()
        pw = new PrintWriter(sw)

        w.writeTo(pw)

        def expected = '''
        

        <script type="text/javascript">
        try {
        
        Genesis
        
        } catch( err ) {}
        </script>
'''
        assertEquals expected, sw.toString()
    }

    void testParsingParensInNestedCurlyBraces() {
        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        def src = '${people.collect {it.firstName.toUpperCase()}}'
        def t = gpte.createTemplate(src, "hello_test")

        def people = [[firstName: 'Peter', lastName: 'Gabriel'], [firstName: 'Phil', lastName: 'Collins']]
        def w = t.make(people: people)

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertEquals "[PETER, PHIL]", sw.toString()
    }

    void testParsingBracketsInNestedCurlyBraces() {
        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        def src = '${people.collect {it.lastName[0]}}'
        def t = gpte.createTemplate(src, "hello_test")

        def people = [[firstName: 'Peter', lastName: 'Gabriel'], [firstName: 'Phil', lastName: 'Collins']]
        def w = t.make(people: people)

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertEquals "[G, C]", sw.toString()
    }

    void testParsingIfs() {
        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        def src = '''<g:if test="${var=='1' || var=='2'}">hello</g:if>'''

        def t = gpte.createTemplate(src, "if_test")

        def w = t.make(var: '1')

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertEquals "hello", sw.toString()
    }


    void testParsingMultilineQuotes() {
        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        def src = '''${{var=='1'}()?"""start.
This is a multi-line string. } 
end.""":''}
<g:if test="${var=='1' || var=='2'}">hello</g:if>
'''

        def t = gpte.createTemplate(src, "test_multiline_str")

        def w = t.make(var: '1')

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)
        
        def result='''start.
This is a multi-line string. } 
end.
hello
'''

        assertEquals result, sw.toString()
    }
    
    void testParsingMultilineQuotes2() {
        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        def src = '''${{var=='1'}()?\'\'\'start.
This is a multi-line string. }
end.\'\'\':''}
<g:if test="${var=='1' || var=='2'}">hello</g:if>
'''

        def t = gpte.createTemplate(src, "test_multiline_str")

        def w = t.make(var: '1')

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)
        
        def result='''start.
This is a multi-line string. }
end.
hello
'''

        assertEquals result, sw.toString()
    }

    void testGscript() {
        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.grailsApplication = createMockGrailsApplication()
        gpte.afterPropertiesSet()

        def src = '''@{page defaultCodec="HTML"}%{if(var=='1') { out.print('hello') } else { out.print('not_ok') } }%'''

        def t = gpte.createTemplate(src, "gscript_test")

        def w = t.make(var: '1')

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertEquals "hello", sw.toString()
    }

    
    void testGRAILS8218() {
        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        
        gpte.afterPropertiesSet()

        def src = '''<g:if test='[pwd:"${actionName}-xx"]'>ok</g:if>'''

        def t = gpte.createTemplate(src, "testGRAILS8218")

        def w = t.make(actionName: 'hello')

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertEquals "ok", sw.toString()
    }
    
    void testGRAILS8199() {
        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        
        gpte.afterPropertiesSet()

        def src = '''<div id='${map["${id}_postfix"]}'/>'''

        def t = gpte.createTemplate(src, "testGRAILS8199")

        def w = t.make(id: 'id', map:[id_postfix:'hello'])

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertEquals "<div id='hello'/>", sw.toString()
    }
    
    void testParsingQuotes() {
        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        def src = '''<g:if test="${var=="1" || var=="2" || var=='}' || var=="{" || var=='"' || var=="\\"" || var=='' || var=="" }">hello</g:if>'''

        def t = gpte.createTemplate(src, "if_test")

        def w = t.make(var: '1')
        def sw = new StringWriter()
        def pw = new PrintWriter(sw)
        w.writeTo(pw)

        assertEquals "hello", sw.toString()

        w = t.make(var: '"')
        sw = new StringWriter()
        pw = new PrintWriter(sw)
        w.writeTo(pw)
        assertEquals "hello", sw.toString()
    }

    void testCreateTemplateWithBinding() {

        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        def t = gpte.createTemplate('Hello ${foo}', "hello_test")
        def w = t.make(foo:"World")

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertEquals "Hello World", sw.toString()
    }

    void testInlineScriptWithValidUnmatchedBrackets() {
        
        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        def t = gpte.createTemplate('''
<% if(true) { %>
Hello ${foo}
<% } %>
<% if(false) { %>
never
<% } else { %>
  
<% } %>
''', "hello_test")
        def w = t.make(foo:"World")

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertEquals "Hello World", sw.toString().trim()
    }
        
    void testInlineScriptWithValidUnmatchedBracketsGspSyntax() {
        
        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        def t = gpte.createTemplate('''
%{ if(true) { }%
Hello ${foo}
%{ } }%
%{ if(false) { }%
never
%{ } else { }%
  
%{ } }%
''', "hello_test")
        def w = t.make(foo:"World")

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertEquals "Hello World", sw.toString().trim()
    }
    
    void testCreateTemplateFromText() {

        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        def t = gpte.createTemplate("<%='hello'%>", "hello_test")
        def w = t.make()

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)

        assertEquals "hello", sw.toString()
    }

    void testForEachInProductionMode() {
        System.setProperty("grails.env", "production")
        GrailsWebUtil.bindMockWebRequest()

        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        def t = gpte.createTemplate("<g:each var='num' in='\${1..5}'>\${num} </g:each>", "foreach_test")
        def w = t.make()

        def sw = new StringWriter()
        def pw = new PrintWriter(sw)

        w.writeTo(pw)
        System.setProperty("grails.env", "development")

        assertEquals "1 2 3 4 5 ", sw.toString()
    }

    void testGetUriWithinGrailsViews() {
        def gpte = new GroovyPagesTemplateEngine(new MockServletContext())
        gpte.afterPropertiesSet()

        assertEquals "/WEB-INF/grails-app/views/myview.gsp", gpte.getUriWithinGrailsViews("/myview")
        assertEquals "/WEB-INF/grails-app/views/myview.gsp", gpte.getUriWithinGrailsViews("myview")
        assertEquals "/WEB-INF/grails-app/views/mydir/myview.gsp", gpte.getUriWithinGrailsViews("mydir/myview")
        assertEquals "/WEB-INF/grails-app/views/mydir/myview.gsp", gpte.getUriWithinGrailsViews("/mydir/myview")
    }

    void tearDown() {
        RequestContextHolder.setRequestAttributes(null)
    }

    void setUp() {
        RequestContextHolder.setRequestAttributes(null)
    }
}
